import numpy as np
from numpy.random import randint, gamma
from sys import stdout

# Generates patterns from nC chains. Each chain has active and inactive phases.
# An active phase has duration tau_on ~ Gamma(w,s), and is followed by an 
# inactive phase of duration tau_off ~ Gamma(w',s'). With the activity onset of a chain,
# a location k is drawn, respecting an non-inhibition structure W_mask
# (W_mask = 0 between inhibited locations). Then one out of nT templates is
# drawn for this location.
# The entire procedure runs nP-times parallelized.
# Resulting pattern data structure: P.shape = (T,nP,K); P.dtype = int
# with 0 indicating the default hypo.
#
# WARNING: Ensure that for any activity constellation a chain will find a new location
#          for becoming active!
#
# In a second step, the chain is "filled" with rate patterns.
# # # #

# output file name

outfname = "./data/template-diri_1e7x7x3x126.npz"
# base params
T = 8e6        # num time steps
nP = 1          # parallelization
nC = 3          # num chains
K = 7           # num locations
nT = 3          # num templates per location

# extended params
seed = 0        # random seed
r_inh = 1       # range of inh (in z_k space)
draw_on = lambda: int(round(gamma(10,20)))
draw_off = lambda: int(round(gamma(10,10)))

# params for filling the activity patterns
local_feature = np.load('data/rate-diri03_7x3x36.npy')
left = lambda k:  (6*(-1 + 3 * k)) % N      # k is location: neuron idx / 3 = k
I = local_feature.shape[2]
N = 7*3*6


# # # # 
# # MAKE CHAINS
# # # #

print " > Generate chains."

# build W_mask
W_mask = np.ones((K,K), dtype=int)
W_mask -= np.eye(K)
for r in xrange(r_inh):
    r += 1
    W_mask -= np.roll(np.eye(K), r, axis=0)
    W_mask -= np.roll(np.eye(K), -r, axis=0)

# func
def draw_location_and_template(p, W_mask):
    idx = W_mask[p.nonzero()].all(0).nonzero()[0]   # allowed locations
    k,i = idx[randint(len(idx))], randint(1,nT+1)   # draw location and template
    return k,i


def add_chain(P): # add a single parallelization
    t_event = np.array([draw_off() for c in xrange(nC)])
    t = t_event.min()       # when
    c = t_event.argmin()    # who
    while t < T:
        dt = draw_on()      # place element for chain c
        k,i = draw_location_and_template(P[t], W_mask)
        P[t:t+dt,k] = i
        t_event[c] = t + dt + draw_off()  # next event in chain c
        t = t_event.min()       # when
        c = t_event.argmin()    # who


np.random.seed(seed)

P = np.zeros((T,nP,K), dtype=int)
for j in xrange(nP):
    add_chain(P[:,j])



# # # # 
# # FILL CHAINS
# # # #

print " > Fill chains."

pattern = P
T = pattern.shape[0]
X = np.zeros((T,nP,N), dtype=np.float16)

for t,p in zip(xrange(T), pattern):
  if t % 10000 == 0: print "\b.",; stdout.flush()
  for j in xrange(nP):
    for k in p[j].nonzero()[0]:
        i = p[j,k]
        x = np.zeros(N)
        x[:I] = local_feature[k,i-1]
        X[t] += np.roll(x, left(k))

if outfname:
    print "\n > Save result."
    np.savez_compressed(outfname, template=X)




