import numpy as np
from sys import stdout

norms = {'2-norm' : lambda r: np.sqrt(np.dot(r,r)), 
         'max-norm' : lambda r: max(abs(r)),
         'abs-x' : lambda r: abs(r[0])
         }


def dist_pbc(x,y,dims,norm=lambda r:np.sqrt(np.dot(r,r))):
    x = np.array(x).astype(float)
    y = np.array(y).astype(float)
    dims = np.array(dims).astype(float)
    r = (x-y) - dims * np.round((x-y)/dims)
    d = norm(r)
    return d

def dist_flat(x,y,dims,norm=lambda r:np.sqrt(np.dot(r,r))):
    x = np.array(x).astype(float)
    y = np.array(y).astype(float)
    dims = np.array(dims).astype(float)
    r = (x-y)
    d = norm(r)
    return d


class NetworkBuilder(object):
    def __init__ (self, K, N, dims, pbc=True):
        self.dims = dims
        self.K = K
        self.N = N
        self.pos_yi = []
        self.pos_zk = []
        self.V = np.zeros((K,N))
        self.W = np.zeros((K,K))
        self.b = np.zeros(K)
        self.V_mask = np.zeros((K,N), dtype=bool)
        self.W_mask = np.zeros((K,K), dtype=bool)
        self.isSet = []
        if pbc:
            self.topo_func = dist_pbc
        else:
            self.topo_func = dist_flat


    def set_pos_raster(self, which, dist=(1,1), offset=(0,0)):
        print " > Network Builder: Set position of " + which + "."
        K = dict(yi=self.N, zk=self.K)[which]
        pos = dict(yi=self.pos_yi, zk=self.pos_zk)[which]
        del pos[:]  # clear list
        idxX = np.arange(offset[0],self.dims[0],dist[0])
        idxY = np.arange(offset[1],self.dims[1],dist[1])
        rasterdims = (len(idxX), len(idxY))
        for x in idxX:
            for y in idxY:
                pos.append((x,y))
        assert len(pos) == K, "NETWORK ERROR: Number of " + which + " neurons (%d) does not match raster (%d)!" % (K, len(pos))
        self.isSet.append(dict(yi="pos_yi", zk="pos_zk")[which])
        return rasterdims

    def set_V(self, connradius, norm="max-norm", p_plast=1.0, Vinit=0.0):
        assert "pos_zk" in self.isSet
        assert "pos_yi" in self.isSet
        norm = norms[norm]
        print " > Network Builder: Set V matrix:",
        numdel = 0
        for k,zpos in enumerate(self.pos_zk):
            s = "%3d" % (round(100 * float(k)/self.K)) + "%"
            print "\b" * numdel + s,; stdout.flush()
            numdel = len(s) + 1
            for i,ypos in enumerate(self.pos_yi):
                if self.topo_func(zpos, ypos, self.dims, norm) <= connradius:
                    if p_plast == 1. or np.random.rand() < p_plast:
                        self.V[k,i] = Vinit
                        self.V_mask[k,i] = 1
        print ""; stdout.flush()
        self.isSet.append("V")
        self.isSet.append("V_mask")


    def set_W(self, b_init, W_inh, connradius, norm='max-norm', p_plast=1., **kwargs):
        assert "V" in self.isSet, "NETWORK ERROR: Input connections not set!"
        assert "V_mask" in self.isSet,"NETWORK ERROR: Input connections not set!"
        norm = norms[norm]
        print " > Network Builder: Set W matrix and b vector:",
        numdel = 0
        for k,zkpos in enumerate(self.pos_zk):
            s = "%3d" % (round(100 * float(k**2)/self.K**2)) + "%"
            print "\b" * numdel + s,; stdout.flush()
            numdel = len(s) + 1
            for l,zlpos in enumerate(self.pos_zk[:k+1]):
                if k==l:
                    self.b[k] = b_init
                elif (self.V_mask[k] * self.V_mask[l]).sum() > 0: 
                    self.W[k,l] = self.W[l,k] = W_inh
                elif self.topo_func( zkpos, zlpos, self.dims, norm) <= connradius:
                    self.W_mask[k,l] = self.W_mask[l,k] = np.random.rand() < p_plast
        print ""; stdout.flush()
        self.isSet.append("W")
        self.isSet.append("b")
        self.isSet.append("W_mask")


    def add_subnet(self, num_net, w, **kwargs):
        assert "W_mask" in self.isSet, "NETWORK ERROR: W_mask is missing."
        idx = np.arange(self.K)
        np.random.shuffle(idx)
        nPerNet = self.K / num_net
        subnet = [np.sort(idx[n*nPerNet:(n+1)*nPerNet]) for n in xrange(num_net)]
        print " > Network Builder: Add subnet structure to W matrix."
        for n in xrange(num_net):
            idx = subnet[n]
            for i,k in enumerate(idx[:-1]):
                for l in idx[i+1:]:
                    if self.W_mask[k,l] == 1:
                        self.W[k,l] = self.W[l,k] = w
        return subnet


    def check(self):
        setlist = ("pos_yi", "pos_zk", "V", "W", "b", "V_mask", "W_mask")
        for s in setlist:
            assert s in self.isSet, "NETWORK ERROR: " + s + "has not been set!"
        return True


if __name__ == "__main__":
    K = 225
    N = 900
    dims = (30,30)
    net = NetworkBuilder(K, N, dims)
    net.set_pos_raster("zk", dist=(2,2), offset=(1,1))
    net.set_pos_raster("yi", dist=(1,1), offset=(0,0))
    net.set_V(connradius=3., norm="max-norm", p_plast=1.0, Vinit=1.0)
    net.set_J(pi_0=1., J_inh=-5., connradius=10., norm='max-norm')



