import numpy as np 


class FromTemplatePoissonGenerator(object):
    """
    Generates poisson inputs from template patterns.
        template : shape = (num_templates, num_inputs), contains expected values of poisson variables
        num_parallel : level of parallelization
        spiking : True or False
        only if spiking:
            tau : num timesteps per psp
            duration : num timesteps to present one pattern

    func generate(T=1, sequence=None):
      if sequence is None:
        Draws a random pattern i from template and presents it for 1 (non-spiking) or duration (spiking) time steps.
        shape = (T, num_parallel, num_inputs)
      elif sequence of shape (T,num_parallel,N):
        use sequence as basis for generation, where sequence[t,n,i] is the template index for 
        neuron i in sampler n at time step t
      if spiking:
          returns template index, spike pattern of length T for num_parallel samplers (with poisson spike trains)
      elif non-spiking:
          returns template index, state pattern of length T for num_parallel samplers (with poisson distributed states)
    """
    def __init__(self, template, num_parallel, spiking, tau=10, duration=200):
        self.template = template / (spiking and tau or 1)
        self.num_parallel = num_parallel
        self.spiking = spiking
        if not spiking:
            tau = 1
            duration = 1
        self.tau = tau
        self.duration = duration

    def generate(self, T=1, sequence=None):
        num_template, N = self.template.shape
        if sequence is None:
            tidx = np.random.randint(0, num_template, (T/self.duration + 1, self.num_parallel))
            tidx = np.repeat(tidx, self.duration, axis=0)[:T]
        elif sequence.shape == (T,self.num_parallel):
            tidx = sequence
        else:
            raise Exception, "Shape of sequence (%s) is inappropriate!" % (str(sequence.shape))
        L = np.array([self.template[tidx[t]] for t in xrange(T)])
        S = np.random.poisson(L)
        return tidx, S


class FromTemplateBernoulliGenerator(object):
    """
    Generates Bernoulli distributed inputs from template patterns
        If spiking: assumes that the state interpretation is "y=1 iff spike within tau".
                    The spike trains are Poisson spike trains, i.e. memory-less.
        template : shape = (num_templates, num_inputs), contains expected values of Bernoulli variables
        num_parallel : level of parallelization
        spiking : True or False
        only if spiking:
            tau : num timesteps per psp
            duration : num timesteps to present one pattern

    func generate(T=1, sequence=None):
      if sequence is None:
        Draws a random pattern i from template and presents it for 1 (non-spiking) or duration (spiking) time steps.
        shape = (T, num_parallel, num_inputs)
      elif sequence of shape (T,num_parallel,N):
        use sequence as basis for generation, where sequence[t,n,i] is the template index for 
        neuron i in sampler n at time step t
      if spiking:
          returns template index, spike pattern of length T for num_parallel samplers (with poisson spike trains)
      elif non-spiking:
          returns template index, state pattern of length T for num_parallel samplers (with poisson distributed states)
    """
    def __init__(self, template, num_parallel, spiking, tau=10, duration=200):
        self.num_parallel = num_parallel
        self.spiking = spiking
        if not spiking:
            tau = 1
            duration = 1
        self.template = (1. - (1. - template)**(1./tau)).astype(np.float16)
        self.tau = tau
        self.duration = duration

    def generate(self, T=1, sequence=None):
        num_template, N = self.template.shape
        print "   > draw tidx"
        if sequence is None:
            tidx = np.random.randint(0, num_template, (T/self.duration + 1, self.num_parallel))
            tidx = np.repeat(tidx, self.duration, axis=0)[:T].astype(np.int32)
        elif sequence.shape == (T,self.num_parallel):
            tidx = sequence
        else:
            raise Exception, "Shape of sequence (%s) is inappropriate!" % (str(sequence.shape))
        print "   > build P"
        P = np.array([self.template[tidx[t]] for t in xrange(T)])
        #P = self.template[tidx]                # avoid! evil numpy bug for large indices
        print "   > draw spikes"
        R = np.random.rand(*P.shape)
        S = R < P
        #np.less(np.random.rand(*P.shape),P,P)   # use P also as spike array to save memory // but: segfault!
        print "   > done"
        return tidx, S



class FromFileLoader(object):
    """
    Reads spike trains from a file. File can be '.npy' or '.npz'. If the file is shorter 
    than the requested input duration, it will be repeatedly presented.

        num_parallel : play the first [:num_parallel] spike streams
        fname_spike : spike file to load
                      if .npz: save spikes as ["S"]
        feat_key  : key of temporal feature idx underlying the spikes (defaults to None; only for '.npz')
        spiking, tau : ignored standard params

        generate(T)
        T : present T time steps (in a loop if T is too large)
        tstart : set a start index; continue presentation if None
        returns featvec, spikearray
    """
    def __init__(self, num_parallel, fname_spike, feat_key=None, spiking=None, tau=None):
        print " > Load input from file '%s'." % fname_spike
        self.num_parallel = num_parallel
        self.fname_spike = fname_spike
        self.feat_key = feat_key
        if fname_spike[-3:] == 'npy':
            self.S = np.load(fname_spike)
            self.feat_key = None
            self.F = None
        elif fname_spike[-3:] == 'npz':
            a = np.load(fname_spike)
            self.S = a['S']
            if feat_key is not None:
                self.F = a[feat_key]
            else:
                self.F = None
        else:
            raise Exception
        self.Tmax, self.nPmax, self.N = self.S.shape
        assert num_parallel <= self.nPmax
        self.nextt = 0

    def generate(self, T=1, tstart=None):
        if tstart is not None:
            self.nextt = tstart % self.Tmax
        s = np.zeros((T,self.num_parallel,self.N), dtype=bool)
        if self.F is not None:
            f = np.zeros((T,self.num_parallel)+self.F.shape[2:], dtype=self.F.dtype)
        else:
            f = None
        idx = np.arange(self.nextt, self.nextt + T) % self.Tmax
        s[:] = self.S[:,:self.num_parallel][idx]
        if self.F is not None:
            f[:] = self.F[:,:self.num_parallel][idx]
        self.nextt = (self.nextt + T) % self.Tmax
        return f,s


# VERY SLOW IMPLEMENTATION of convolve2d
#def applyPSPkernel(S, kernel, postprocess=lambda Y: Y):
    #"""
    #Convolve spike pattern S with kernel. Finally apply postprocess.
    #SY : shape = (T,num_parallel,N)
    #kernel : single psp shape, e.g. [1.]*tau for rectangular psps
    #postprocess : function that is applied after convolution,
                  #e.g. lambda Y : np.clip(Y, 0., 1.) for non-additive psps

    #returns convolved state vector Y
    #"""
    #from scipy.signal import convolve2d
    #T,num_parallel,N = S.shape
    #kernel = np.transpose([kernel]) # make kernel suitable for convolve2d
    #Y = convolve2d(S.reshape(T,num_parallel*N), kernel)[:T].reshape(T,num_parallel,N)
    #Y = postprocess(Y)
    #return Y


def applyPSPkernel(S, kernel, postprocess=lambda Y: Y, dtype=np.float):
    """
    Convolve spike pattern S with kernel. Finally apply postprocess.
    SY : shape = (T,num_parallel,N)
    kernel : single psp shape, e.g. [1.]*tau for rectangular psps
    postprocess : function that is applied after convolution,
                  e.g. lambda Y : np.clip(Y, 0., 1.) for non-additive psps

    returns convolved state vector Y
    """
    T,num_parallel,N = S.shape
    #kernel = np.transpose([kernel]) # make kernel suitable for convolve2d
    S = S.reshape(T,num_parallel*N)
    Y = np.zeros((T,num_parallel*N), dtype=dtype)
    for i in xrange(num_parallel*N):
        Y[:,i] = np.convolve(S[:,i], kernel)[:T]
    #Y = convolve2d(, kernel)[:T]
    S = S.reshape(T,num_parallel,N)
    Y = Y.reshape(T,num_parallel,N)
    Y = postprocess(Y)
    return Y



def applyFastBinaryKernel(S, tau, outdtype=np.float16):
    #Y = np.zeros(S.shape, dtype=outdtype)
    Y = S.copy()
    #Y[:] = S
    for i in xrange(1, tau):
        np.logical_or(Y[i:], S[:-i], out=Y[i:])
    return Y.astype(outdtype)


