import numpy as np
from sys import stdout, platform
from multiprocessing import Process, Value, Array, Event, Semaphore
import shmarray

class Coordinator(object):
    """
    Initializes parallel samplers and coordinates the sampling process.
        sampler: instantiation of a sampler. 
                 WARNING: For recurrent learning, sampler must have a shmarray-recurrent weight matrix!
                 WARNING: For plastic biases, sampler must have a shmarray-bias vector!
        num_parallel: int or 'auto'.
        spiking: Bool whether spikes should be obtained from the samplers.
    """
    def __init__(self, sampler, num_parallel, spiking):
        print "\n > Init Coordinator."
        self.sampler = sampler
        self.K = sampler.K
        self.spiking = spiking
        # Estimate num of processors
        if num_parallel == 'auto':
            num_parallel = get_num_processors() - 1
            num_parallel = max(num_parallel, 1)
        print " > Level of parallelization: %d (= %d processes)" % (num_parallel, num_parallel + 1)
        self.num_parallel = num_parallel
        # communication and synchronization related arrays, Events,...
        self.startLoopEv = Event()               # the exit of the gate
        self.endLoopEv = Event()                 # the entrance of the gate 
        self.samplersAtEntrance = Semaphore(0)   # counter: num workers inside the gate
        # shared memory
        self.ext = [shmarray.zeros(self.K, dtype='d') for i in xrange(self.num_parallel)]
        self.z_collect = [shmarray.zeros(self.K, dtype='i') for i in xrange(self.num_parallel)]
        if self.spiking:
            self.s_collect = [shmarray.zeros(self.K, dtype='i') for i in xrange(self.num_parallel)]  # spikes
        self.samplerIsReady = Array('i', self.num_parallel)   # if sample is available
        self.terminate = Value('i', 0)
        # arrays of the master
        self.Z = np.zeros((self.num_parallel, self.K), dtype=int)       # for collection and return
        if self.spiking:
            self.S = np.zeros((self.num_parallel, self.K), dtype=int)   # for collection and return
        self.sampleCount = np.zeros(self.num_parallel, dtype=int)
        self.obtained = np.zeros(self.num_parallel, dtype=int)
        self.obtainNow = np.zeros(self.num_parallel, dtype=int)
        # start parallel processes
        kwargs = dict(sampler=self.sampler,                        # args shared by all
                      startLoopEv=self.startLoopEv,
                      endLoopEv=self.endLoopEv, 
                      samplersAtEntrance=self.samplersAtEntrance, 
                      samplerIsReady=self.samplerIsReady, 
                      terminate=self.terminate)
        if self.spiking:
            self.processes = [MyProcess(idx=i, z=self.z_collect[i], s=self.s_collect[i], ext=self.ext[i], **kwargs) for i in xrange(self.num_parallel)]
        else:
            self.processes = [MyProcess(idx=i, z=self.z_collect[i], s=None, ext=self.ext[i], **kwargs) for i in xrange(self.num_parallel)]

    def start_samplers(self):
        # start the processes
        print " > Starting %d samplers." % self.num_parallel
        for p in self.processes:
            p.start()

    def distribute_ext(self, ext):
        for i in xrange(self.num_parallel):
            self.ext[i][:] = ext[i]

    def sampling_initiate(self):
        self.startLoopEv.clear()                         # lock gate exit
        self.endLoopEv.set()                             # let them into the gate
        while self.samplersAtEntrance.get_value() != self.num_parallel: # wait for s. to be in the gate
            pass
        self.endLoopEv.clear()                           # lock gate entrance
        self.startLoopEv.set()                           # let them into their loops

    def sampling_collect(self):
        while not self.obtained.all():   # obtain data from all samplers
            self.obtainNow = np.logical_not(self.obtained) * np.array(self.samplerIsReady) # now=(not yet done) * ready samplers
            for i in self.obtainNow.nonzero()[0]:
                self.Z[i] = self.z_collect[i]             # fetch the sample
                if self.spiking:
                    self.S[i] = self.s_collect[i]        # and maybe spikes
                self.sampleCount[i] += 1
            np.logical_or(self.obtained, self.obtainNow, out=self.obtained)   # update what has been obtained
        self.obtained.fill(0)                            # clear what has been done
        self.samplerIsReady[:] = [0] * self.num_parallel      # no sampler has a sample any more
        if self.spiking:
            return self.Z, self.S
        else:
            return self.Z

    def finalize(self):
        print "\n > Terminating samplers."
        print " > Received samples per process:", self.sampleCount
        self.terminate.value = 1                         # tell the workers to exit the loop
        self.endLoopEv.set()                             # open the gate so they can reach the end of the loop
        for p in self.processes:          # finish correctly by joining the process (which then directly quits)
            p.join()






def get_num_processors():
    """Taken from a report by Sturla Molden, University of Oslo, 2009."""
    assert platform == 'linux2', "Automatic estimation of processors only works on Linux 2.x!"
    retv = 0
    with open('/proc/cpuinfo', 'rt') as cpuinfo:
        for line in cpuinfo:
            if line[:9] == 'processor': retv += 1
    return retv


class MyProcess(Process):
    def __init__ (self, idx, z, s, ext, sampler, startLoopEv, endLoopEv, samplersAtEntrance, samplerIsReady, terminate):
        Process.__init__(self)
        self.idx = idx                                  # initialize all args
        self.z = z
        self.s = s
        self.ext = ext
        self.sampler = sampler
        self.startLoopEv = startLoopEv
        self.endLoopEv = endLoopEv
        self.samplersAtEntrance = samplersAtEntrance
        self.samplerIsReady = samplerIsReady
        self.terminate = terminate
        self.sampleCount = 0
    def run(self):
        np.random.seed()                                # set a random seed
        #print "Process %s alive." % self.name
        while self.terminate.value == 0:
            self.samplersAtEntrance.release()           # report worker inside the gate
            self.startLoopEv.wait()                     # wait for the exit to open
            self.samplersAtEntrance.acquire()           # report worker outside the gate
            self.sampler.ext[:] = self.ext                  # fetch external input from shared memory
            self.z[:] = self.sampler.advance().astype(int)      # draw sample to shared-memory
            if self.s is not None: # spiking sim
                self.s[:] = self.sampler.zeta == self.sampler.tau  # write spikes to shared memory
            #print "%s got a sample (%d)." % (self.name, self.sampleCount)
            self.sampleCount += 1
            self.samplerIsReady[self.idx] = 1           # report sample is ready
            self.endLoopEv.wait()                       # wait for the entrance to open
        #print "%s done; produced %d samples." % (self.name, self.sampleCount)

