import numpy as np

class SamplerAbsref(object):
    """
    Neural Sampling Network with absolute refractory mechanism.
        W : recurrent weight matrix with W_ij = W_ji and W_ii = 0
        b : neuronal bias vector
        tau : refractory time in simulation time steps
        shuffleIdx : If True, neurons are updated in random order
    advance network by one time step via self.advance()
        returns new (binary) network state self.z
    """
    def __init__(self, W, b, tau, shuffleIdx=True):
        # save the network parameters (and check that all dimensions match)
        self.W = W
        assert (W.diagonal() == 0.).all(), "Weight matrix must have zero diagonal!"
        self.b = b
        self.tau = tau
        assert len(b)==len(W)==len(W[0]), "Dimensions don't match!"
        assert isinstance(tau, int) and (tau > 0), "Tau must be a positive integer!"
        # arrays for internal state variables
        self.K = len(b)                             # number of neurons
        self.idx_order = np.arange(self.K)          # neurons can be updated in any order
        self.shuffleIdx = shuffleIdx
        if shuffleIdx: np.random.shuffle(self.idx_order)
        self.ext = np.zeros(self.K)                 # external input (adds to bias)
        self.z = np.zeros(self.K, dtype=int)        # binary state vector
        self.zeta = np.zeros(self.K, dtype=int)     # internal refractory state vector
        # supporting functions and constants
        self.logTau = np.log(tau)
        self.sigma = lambda x: 1./(1.+np.exp(-x))
        self.spike_in_01 = lambda u: np.random.rand() < self.sigma(u - self.logTau)

    def advance(self):                                                  # advance one MCMC step
        if self.shuffleIdx: np.random.shuffle(self.idx_order)
        for k in self.idx_order:                                       # for each neuron
            if self.zeta[k] > 1:                                        # if refractory:
                self.zeta[k] -= 1                                       # advance internal state,
                continue                                                # and continue to next neuron k+1
            uk = np.dot(self.W[k], self.z) + self.b[k] + self.ext[k]    # otherwise: calculate membrane potential
            self.zeta[k] = self.tau * self.spike_in_01(uk)              # spike or decay to tau=0
            self.z[k] = self.zeta[k] > 0                                # set z[k] appropriately
        return self.z                                                   # return the new network state
 



class SamplerGibbs(object):
    """
    Gibbs Sampler.
        W : recurrent weight matrix with W_ij = W_ji and W_ii = 0
        b : neuronal bias vector
        ext : additional external input vector (adds to b)
        burnin : no of full Gibbs iterations before sample is taken
        shuffleIdx : If True, neurons are updated in random order
    advance network by one time step via self.advance()
        returns new (binary) network state self.z
    """
    def __init__(self, W, b, burnin=1, shuffleIdx=True):
        # save the network parameters (and check that all dimensions match)
        self.W = W
        self.b = b
        self.burnin = burnin
        assert len(b)==len(W)==len(W[0]), "Dimensions don't match!"
        assert isinstance(burnin, int) and (burnin > 0), "Burnin must be a positive integer!"
        # arrays for internal state variables
        self.K = len(b)                             # number of neurons
        self.idx_order = np.arange(self.K)          # neurons can be updated in any order
        if shuffleIdx: np.random.shuffle(self.idx_order)
        self.ext = np.zeros(self.K)                 # external input (adds to bias)
        self.z = np.zeros(self.K, dtype=int)        # binary state vector
        # supporting functions and constants
        self.sigma = lambda x: 1./(1.+np.exp(-x))
        self.resamplek = lambda u: np.random.rand() < self.sigma(u)

    def advance(self):
        for i in xrange(self.burnin):                                       # advance burnin MCMC steps
            for k in self.idx_order:                                        # for each neuron
                uk = np.dot(self.W[k], self.z) + self.b[k] + self.ext[k]    # calculate membrane potential
                self.z[k] = self.resamplek(uk)                              # set z[k] appropriately
        return self.z                                                   # return the new network state
 
