import numpy as np
import itertools

class Synapses (object):
    '''Base class.'''
    classname = "Synapses Base Class"
    __getitem__ = lambda self, key: np.ndarray.__getitem__(self.W, key)
    __setitem__ = lambda self, key, val: np.ndarray.__setitem__(self.W, key, val)

    def __init__(self, K, N, syninit, mask=None):
        self.W = np.zeros((K,N))
        if mask is None:
            mask = np.ones((K,N), dtype=bool)
        self.W_mask = mask
        for k,i in itertools.product(range(K), range(N)):
            self.W[k,i] = self.W_mask[k,i] * syninit(k,i)

    def __repr__(self):
        s = self.__class__.classname + "\n\n" + repr(self.W)
        return s

    def transmit(self, y):
        return np.dot(self.W, y)

    def plasticity(self, z, y):
        return NotImplemented



class PoissonSynapses (Synapses):
    '''For a Poisson mixture model with default hypothesis.

      modelparams: dict(lambda0i=1., semlike=False)
      calcAk: whether Ak is added to input potential (default=True (=no homeostasis))
      Notes:
        * implementation supports vector of lambda0i.
        * semlike: True = .../A"; False = simple gradient rule
        '''
    classname = "Synapses Poisson model w/ default"
    default_modelparams = dict(lambda0i=1., semlike=False, only_exc=False)

    def __init__(self, K, N, syninit, mask=None, calcAk=True, **modelparams):
        Synapses.__init__(self, K, N, syninit, mask)
        self.calcAk = calcAk
        self.param = dict()
        self.param.update(self.__class__.default_modelparams, **modelparams)

    def transmit(self, y):
        if self.calcAk:
            Ak = self.get_Ak()
        else:
            Ak = 0.
        inp = -Ak + np.dot(self.W, y)
        return inp

    def plasticity(self, eta, z, y):
        V0 = np.log(self.param['lambda0i'])
        assert z.ndim == y.ndim
        if z.ndim == 1:
            z = z[None,:]
            y = y[None,:]
        if self.param['semlike']:
            dW = np.exp(-self.W - V0) * np.dot(z.T,y) - z.sum(0)[:,None]
            #dW = z[:,None] * (np.exp(-self.W - V0) * y - 1.)
        else:
            dW = np.dot(z.T,y) - np.exp(self.W + V0) * z.sum(0)[:,None]
            #dW = z[:,None] * (y - np.exp(self.W + V0))
        dW *= eta * self.W_mask
        self.W += dW
        if self.param['only_exc']:
            np.clip(self.W, 0., 1000., out=self.W)
        return dW

    def get_Ak(self):
        '''Return the Ak values based on weight and default hypo'''
        l0 = self.param['lambda0i']
        Aki = l0 * (np.exp(self.W) - 1.)
        Ak = Aki.sum(1)
        return Ak

    def get_centers(self):
        '''Returns the cluster centers.'''
        return self.param['lambda0i'] * np.exp(self.W)

    def get_V0(self):
        '''Return the implicit weights V_0i'''
        return np.log(self.param['lambda0i'])

    def set_V0(self, V0):
        '''Set lambda0i via implicit weights V0'''
        l0 = np.exp(V0)
        self.param['lambda0i'] = l0



class BernoulliSynapses (Synapses):
    '''For a Bernoulli mixture model with default hypothesis.

      modelparams: dict(pi0i=0.5, semlike=False)
      calcAk: whether Ak is added to input potential (default=True (=no homeostasis))
      Notes:
        * implementation supports vector of lambda0i.
        * semlike: True = .../A"; False = simple gradient rule
        '''
    classname = "Synapses Bernoulli model w/ default"
    default_modelparams = dict(pi0i=0.5, semlike=False, only_exc=False)

    def __init__(self, K, N, syninit, mask=None, calcAk=True, **modelparams):
        Synapses.__init__(self, K, N, syninit, mask)
        self.calcAk = calcAk
        self.param = dict()
        self.param.update(self.__class__.default_modelparams, **modelparams)
        self._sigma = lambda x: 1. / ( 1. + np.exp(-x))
        self._logit = lambda x: np.log(x / (1.-x))

    def transmit(self, y):
        if self.calcAk:
            Ak = self.get_Ak()
        else:
            Ak = 0.
        inp = -Ak + np.dot(self.W, y)
        return inp

    def plasticity(self, eta, z, y):
        V0 = self._logit(self.param['pi0i'])
        assert z.ndim == y.ndim
        if z.ndim == 1:
            z = z[None,:]
            y = y[None,:]
        if self.param['semlike']:
            raise Exception, "TODO: devide grad rule by s(x)*s(-x)"
            dW = self._sigma(-self.W - V0)**-1 * np.dot(z.T,y) - z.sum(0)[:,None]
            #dW = z[:,None] * (np.exp(-self.W - V0) * y - 1.)
        else:
            dW = np.dot(z.T,y) - self._sigma(self.W + V0) * z.sum(0)[:,None]
            #dW = z[:,None] * (y - np.exp(self.W + V0))
        dW *= eta * self.W_mask
        self.W += dW
        if self.param['only_exc']:
            np.clip(self.W, 0., 1000., out=self.W)
        return dW

    def get_Ak(self):
        '''Return the Ak values based on weight and default hypo'''
        pi0i = self.param['pi0i']
        V0 = self._logit(pi0i)
        Aki = self.W  - np.log(self._sigma(self.W + V0) / pi0i)
        Ak = Aki.sum(1)
        return Ak

    def get_centers(self):
        '''Returns the cluster centers.'''
        V0 = self._logit(self.param['pi0i'])
        return self._sigma(self.W + V0)

    def set_centers(self, piki):
        '''Set weights based on piki and pi0i'''
        self.W[:] = self.W_mask * (self._logit(piki) - self._logit(self.param['pi0i']))

    def get_V0(self):
        '''Return the implicit weights V_0i'''
        return self._logit(self.param['pi0i'])

    def set_V0(self, V0):
        '''Set pi0i via implicit weights V0'''
        self.param['pi0i'] = self._sigma(V0)






if __name__ == '__main__':
    modelparams = dict(lambda0i=0.1, semlike=True)
    syninit = lambda *args: 0
    mask = np.zeros((2,3), dtype=bool)
    mask[:,:2] = 1
    V = PoissonSynapses(2, 3, syninit, mask=mask, **modelparams)

    z = np.array([0,1])
    for i in xrange(10000):
        V.plasticity(0.001, z, np.random.poisson((0.1, 0.5, 0.9)))
