import numpy as np
from sys import stdout

# # # # # # # # # # # # # #
# # #  INITIALIZATION # # #
# # # # # # # # # # # # # #

DSL = lambda t: '-'.join(("%02d" % i for i in t[:6]))

def get_param_file():
    from sys import argv
    if 'ipython' in argv:
        paramfilename = argv[2]
    else:
        paramfilename = argv[1]
    if paramfilename[-3:] == ".py":
        paramfilename = paramfilename[:-3]
    return paramfilename

def extract_params(param, globals_dict, keep=(), first_level=True):
    """
    Bring all params in param into the current name space. 
    Provide the object dict as globals_dict; e.g. via globals()
    Stop dict-nesting for keys in keep.
    """
    if first_level: print " > Adding params to namespace:",
    for key, value in param.iteritems():
        if isinstance(value, dict) and (key not in keep):
            extract_params(value, globals_dict, keep, first_level=False)
        else:
            print key, ' ',
            globals_dict[key] = value
    if first_level: print '\n'


def rec_update(dst, src):
    '''
    recursively updates a dict, thus values being dicts themselves get updated rather than overwritten. 
    Overwrites dst, so the returnvalue is just for fun...
    '''
    for key in dst.keys():
        if src.has_key(key) and isinstance(dst[key], dict) and  isinstance(src[key], dict):
            rec_update(dst[key], src[key])
        elif src.has_key(key):
            dst[key] = src[key]
    for key in src.keys():
        if not dst.has_key(key):
            dst[key] = src[key]
    return dst



def memorycheck(T, arrays, limit=0.9):
    """
    Checks whether enough memory is available on the machine. Raises Exception if more than limit would be used.
    T     : number of timesteps
    num_parallel: level parallelization
    array : list of arrays to check with
            a = (dims, dtype)
                e.g. dims = (num_parallel,K,N) and dtype="float"

    Returns estimated usage in MB.
    """
    estimate = 0.
    for dims,dt in arrays:
        estimate += T * np.prod(dims) * np.dtype(dt).itemsize / 2.**20
    import os
    freemem = int(os.popen("free -m").readlines()[2].split()[3])
    s = "Estimated memory usage: %dMB of %dMB currently free." % (int(estimate), freemem)
    if estimate > limit * freemem:
        raise Exception("MEMORY ERROR! " + s + " (Limit: %.2f)" % limit)
    print " > " + s
    return estimate


# # # # # # # # # # # # # # # # # #
# # # # # ONLINE STATISTICS # # # #
# # # # # # # # # # # # # # # # # #

class ParameterEvolution(object):
    """
    Gets a **dict of parameter-vectors called P that should be monitored.
    self.get_stats(**P): < v/|v| ,w/|w|> where
        v = p(t)-p(t-x)
        w = p(t-x)-p(t-2x)
        for every parameters vector p,
        and t-x is the time of the last call
    if Nfilter > 1: update second-to-last state only partially.
    """
    def __init__(self, Nfilter=1, **P):
        self.Nfilter = float(Nfilter)
        self.p_names = P.keys()
        self.num_calls = 0
        self.P = P
        self.P1 = dict.fromkeys(self.P, 0)
        self.P2 = dict.fromkeys(self.P, 0)

    def _cos_theta(self, v, w):
        from numpy.linalg import norm
        v = v.flatten() / norm(v.flatten())
        w = w.flatten() / norm(w.flatten())
        return np.dot(v,w)

    def get_stats(self,**P):
        assert P.keys() == self.p_names, "  > ERROR: Invalid keys for param evolution! %s, %s" % (str(P.keys()), str(self.p_names))
        for k in self.p_names:
            self.P2[k] = (1.-1./self.Nfilter) * self.P2[k] + 1./self.Nfilter * self.P1[k]
        self.P1.update(self.P)
        for k in self.p_names:  # make a copy of newly arriving params
            self.P[k] = P[k].copy()
        self.num_calls += 1
        if self.num_calls <=2:
            print "  > No parameter evolution available, yet."
            return
        print "  > Parameter evolution:"
        for k in self.p_names:
            x = self._cos_theta(v=self.P[k]-self.P1[k], w=self.P1[k]-self.P2[k])
            print "    > %s: %.3e" % (k, x)




# # # # # # # # # # # # # # # # # #
# # # DISTRIBUTION EVALUATION # # #
# # # # # # # # # # # # # # # # # #

def joint_from_data(Z, idx=None, priorObservations=0.):
    """
    Extract joint over binary z_k, k in idx, and marginalizes over other z_i, i not in idx
        Z: binary data vector of shape (T,K) with T = #samples and K = #variables
        idx: List of indices; None means all indices
        priorObservations: Assume priorObservations of each state (bias for counting)
    """
    if idx is None:
        idx = np.arange(Z.shape[-1])
    Z = Z[:,idx]
    T,K = Z.shape
    b = [2**i for i in reversed(range(K))]
    N = np.bincount(np.dot(Z,b))
    P = priorObservations * np.ones(2**K)
    P[:len(N)] += N
    P /= P.sum()
    states = [np.binary_repr(n,K) for n in xrange(2**K)]
    return states, P


def joint_of_BM(b, W):
    """
    Analytical calculation of Boltzmann machine.
        b, W: params as usual
    """
    from itertools import product as iterprod
    assert len(b) == len(W)
    assert (W.diagonal() == 0.).all()
    K = len(b)
    E = lambda z: -np.dot(z, 0.5 * np.dot(W,z) + b )
    P = np.zeros(2**K)
    for i,z in enumerate(iterprod(*[[0,1]]*K)):
        P[i] = np.exp(-E(z))
    P /= P.sum()
    states = [np.binary_repr(n,K) for n in xrange(2**K)]
    return states, P


def joint_of_BM_indexed(b, W, Z):
    """
    Analytical calculation of Boltzmann machine,
      for large networks with only a small fraction of allowed states.
        b, W: params as usual
        Z: TxK array of the T allowed states
    """
    assert len(b) == len(W)
    assert len(b) == Z.shape[1]
    assert (W.diagonal() == 0.).all()
    K = len(b)
    E = lambda z: -np.dot(z, 0.5 * np.dot(W,z) + b )
    P = np.zeros(len(Z))
    for i,z in enumerate(Z):
        P[i] = np.exp(-E(z))
    P /= P.sum()
    states = [ ''.join([str(int(zk)) for zk in z])  for z in Z]
    return states, P


def calc_marginals(P,K):
    P = P.reshape([2]*K)
    p = np.zeros(K)
    for k in xrange(K):
        p[k] = P.swapaxes(0,k)[1].sum()
    return p


def calc_marginals_indexed(P,Z):
    K = Z.shape[1]
    p = np.zeros(K)
    for k in xrange(K):
        idx = Z[:,k] == 1
        p[k] = P[idx].sum()
    return p

KL = lambda p,q: np.log( (p/q)**p ).sum()           # stable for lim_p(z) -> 0



# # # # # # # # # # 
# # # LOGGER # # #
# # # # # # # # # #

class Logger(object):
    """
    Keeps track of transients of minor importance.
    These are saved every SAVEEVERY time steps via
        Logger.track(t)
    Provides an API for writing results to disk:
        Logger.save(individual=("A", "B"), archive=("C", "D"))
    "A", "B", "C", "D" can be either global objects or members of Logger.
    "A", "B" will be saved to individual files, "C", "D" in a joint compressed archive.
    **kwargs are: name = ("globalname", dims, "dtype"), e.g. Vt = ("V", (K,N), 'float')
    """
    def __init__(self, DSL, T, SAVEEVERY, TEST, globalsdict, **kwargs):
        self.DSL = DSL
        self.T = T
        self.SAVEEVERY = SAVEEVERY
        self.TEST = TEST
        if TEST:
            print " > WARNING: TESTRUN! No results will be saved!"
        if not TEST:
            from os import mkdir
            dirname = "./out/" + self.DSL
            mkdir(dirname)
        self.fname = lambda s: "./out/" + self.DSL + "/" + s
        self.logtimes = range(0, T, SAVEEVERY)
        self.glob = globalsdict
        self.logdict = dict()
        for key, (name,dims,dtype) in kwargs.iteritems():
            self.logdict[key] = name
            self.__setattr__(key, np.zeros((len(self.logtimes),)+dims, dtype=dtype) )

    def save_param(self, paramfilename):
        if not self.TEST:
            import shutil
            shutil.copy(paramfilename+'.py', self.fname("param")+".py")

    def track(self, t):
        if t in self.logtimes:
            i = self.logtimes.index(t)
            for key,name in self.logdict.iteritems():
                getattr(self, key)[i] = self.glob[name][:]
                #np.ndarray.__setitem__(self.__getattribute__(key), i, self.glob[name])

    def save(self, individual, archive):
        print ""
        if self.TEST:
            print " > Skip writing to file (testrun)!"
            return
        print " > Save results. DSL is", self.DSL, "\n >",; stdout.flush()
        saveindiv = lambda s,a: isinstance(s, str) and np.save(self.fname(s), a[:])
        for s in individual:
            print s,"\b,",; stdout.flush()
            if hasattr(self, s):
                a = self.__getattribute__(s)
            else:
                a = self.glob[s]
            saveindiv(s, a)
        kwargs = dict()
        for s in archive:
            if hasattr(self, s):
                kwargs[s] = self.__getattribute__(s)
            else:
                kwargs[s] = self.glob[s][:]
        print "Archive:", archive,; stdout.flush()
        try:
            np.savez_compressed(self.fname("archive"), **kwargs)
            print "\b."
        except e:
            print "\n > WARNING: Saving of archive failed!"
            print e






