import numpy as np
import pylab as pl
from BaseClassPanel import BaseClassPanel

class RecurrentConnectionsPanel(BaseClassPanel):
    def setup(self, W_mask, dims, maxconn=50, seed=0, legend=True):
        self.W_mask = W_mask
        self.dims = dims
        self.maxconn = maxconn
        self.seed = seed
        self.legend = legend
        self.ax.set_aspect('equal')

    def update(self, W, title=''):
        # check data types
        if not isinstance(W, np.ndarray):
            W = np.array(W)
        assert W.ndim == 2, "W must be two dimensional!"
        dims = self.dims
        K = np.prod(dims)
        assert W.shape == (K,K), "W does not match dims!"
        assert W.shape == self.W_mask.shape, "W does not match W_mask!"
        ax = self.ax
        ax.clear()
        # add grid
        Y,X = np.meshgrid(np.arange(dims[0]), np.arange(dims[1]))
        ax.plot(X.flatten(), Y.flatten(), '.', ms=10, c='0.3')
        # get connection indices
        ks,js = (self.W_mask * np.tri(K,k=-1)).nonzero()
        idx = np.arange(len(ks))
        rng_state = np.random.get_state()
        np.random.seed(self.seed)
        np.random.shuffle(idx)
        np.random.set_state(rng_state)
        idx = idx[:self.maxconn]
        ks, js = ks[idx], js[idx]
        # plotting params
        W_mask = self.W_mask
        wmax_exc = (W[W_mask] > 0).any() * W[W_mask].max()
        wmax_inh = (W[W_mask] < 0).any() * W[W_mask].min()
        eps = 0.00001
        lc_exc = lambda w, wmax: np.array([0.5,0.5,0.5]) + np.array([1.,-1.,-1.]) * 0.5 * w / (wmax+eps)
        lc_inh = lambda w, wmax: np.array([0.5,0.5,0.5]) + np.array([-1.,-1.,1.]) * 0.5 * w / (wmax-eps)
        lw = lambda w, wmax: (5. - 0.5) * w / (wmax +  2 * ((w >= 0) - 0.5) * eps) + 0.5
        # plot
        for k,j in zip(ks,js):
            w = W[k,j]
            if w >= 0:
                wmax = wmax_exc
                lc = lc_exc
            else:
                wmax = wmax_inh
                lc = lc_inh
            y,x = np.unravel_index([k,j], dims)
            ax.plot(x, y, ls='-', c=lc(w, wmax), lw=lw(w, wmax))
        if self.legend:
            # red line
            x0,y0 = 0, -0.5
            x1,y1 = x0 + 0.15 * (dims[1] - 1), y0
            ax.plot((x0,x1), (y0,y1), ls='-', c=lc_exc(wmax_exc, wmax_exc), lw=lw(wmax_exc, wmax_exc))
            x,y = x1 + 0.07 * (dims[1] - 1), y0
            ax.text(x, y, "%.3f" % wmax_exc, ha='left', va='center')
            # blue line
            x0,y0 = 0.5 * (dims[1] - 1), y0
            x1,y1 = x0 + 0.15 * (dims[1] - 1), y0
            ax.plot((x0,x1), (y0,y1), ls='-', c=lc_inh(wmax_inh, wmax_inh), lw=lw(wmax_inh, wmax_inh))
            x,y = x1 + 0.07 * (dims[1] - 1), y0
            ax.text(x, y, "%.3f" % wmax_inh, ha='left', va='center')
        # finalize
        ax.set_xlabel('')
        ax.set_ylabel('')
        ax.set_xticks([]) # (np.arange(0,dims[1],2))
        ax.set_yticks([]) # (np.arange(0,dims[0],2))
        ax.set_frame_on(False)
        ax.set_xlim(-0.25,dims[1]-0.75)
        ax.set_ylim(-0.25-0.5*self.legend,dims[0]-0.75)
        ax.set_title(title)


